package adapter

import (
	"context"
	"database/sql"
	"errors"
	"fmt"

	"gitea.dwysokinski.me/twhelp/corev3/internal/bun/bunmodel"
	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"github.com/uptrace/bun"
)

type EnnoblementBunRepository struct {
	db bun.IDB
}

func NewEnnoblementBunRepository(db bun.IDB) *EnnoblementBunRepository {
	return &EnnoblementBunRepository{db: db}
}

func (repo *EnnoblementBunRepository) Create(ctx context.Context, params ...domain.CreateEnnoblementParams) error {
	if len(params) == 0 {
		return nil
	}

	ennoblements := make(bunmodel.Ennoblements, 0, len(params))

	for _, p := range params {
		base := p.Base()
		ennoblements = append(ennoblements, bunmodel.Ennoblement{
			ServerKey:  p.ServerKey(),
			VillageID:  base.VillageID(),
			NewOwnerID: base.NewOwnerID(),
			NewTribeID: base.NewTribeID(),
			OldOwnerID: base.OldOwnerID(),
			OldTribeID: base.OldTribeID(),
			Points:     base.Points(),
			CreatedAt:  base.CreatedAt(),
		})
	}

	if _, err := repo.db.NewInsert().
		Model(&ennoblements).
		Ignore().
		Returning("").
		Exec(ctx); err != nil {
		return fmt.Errorf("something went wrong while inserting ennoblements into the db: %w", err)
	}

	return nil
}

func (repo *EnnoblementBunRepository) List(
	ctx context.Context,
	params domain.ListEnnoblementsParams,
) (domain.ListEnnoblementsResult, error) {
	var ennoblements bunmodel.Ennoblements

	if err := repo.db.NewSelect().
		Model(&ennoblements).
		Apply(listEnnoblementsParamsApplier{params: params}.apply).
		Scan(ctx); err != nil && !errors.Is(err, sql.ErrNoRows) {
		return domain.ListEnnoblementsResult{}, fmt.Errorf("couldn't select ennoblements from the db: %w", err)
	}

	converted, err := ennoblements.ToDomain()
	if err != nil {
		return domain.ListEnnoblementsResult{}, err
	}

	return domain.NewListEnnoblementsResult(separateListResultAndNext(converted, params.Limit()))
}

func (repo *EnnoblementBunRepository) ListWithRelations(
	ctx context.Context,
	params domain.ListEnnoblementsParams,
) (domain.ListEnnoblementsWithRelationsResult, error) {
	var ennoblements bunmodel.Ennoblements

	if err := repo.db.NewSelect().
		Model(&ennoblements).
		Apply(listEnnoblementsParamsApplier{params: params}.apply).
		Relation("Village", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(bunmodel.VillageMetaColumns...)
		}).
		Relation("NewOwner", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(bunmodel.PlayerMetaColumns...)
		}).
		Relation("NewTribe", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(bunmodel.TribeMetaColumns...)
		}).
		Relation("OldOwner", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(bunmodel.PlayerMetaColumns...)
		}).
		Relation("OldTribe", func(q *bun.SelectQuery) *bun.SelectQuery {
			return q.Column(bunmodel.TribeMetaColumns...)
		}).
		Scan(ctx); err != nil && !errors.Is(err, sql.ErrNoRows) {
		return domain.ListEnnoblementsWithRelationsResult{}, fmt.Errorf("couldn't select ennoblements from the db: %w", err)
	}

	converted, err := ennoblements.ToDomainWithRelations()
	if err != nil {
		return domain.ListEnnoblementsWithRelationsResult{}, err
	}

	return domain.NewListEnnoblementsWithRelationsResult(separateListResultAndNext(converted, params.Limit()))
}

type listEnnoblementsParamsApplier struct {
	params domain.ListEnnoblementsParams
}

//nolint:gocyclo
func (a listEnnoblementsParamsApplier) apply(q *bun.SelectQuery) *bun.SelectQuery {
	if serverKeys := a.params.ServerKeys(); len(serverKeys) > 0 {
		q = q.Where("ennoblement.server_key IN (?)", bun.In(serverKeys))
	}

	for _, s := range a.params.Sort() {
		switch s {
		case domain.EnnoblementSortCreatedAtASC:
			q = q.Order("ennoblement.created_at ASC")
		case domain.EnnoblementSortCreatedAtDESC:
			q = q.Order("ennoblement.created_at DESC")
		case domain.EnnoblementSortIDASC:
			q = q.Order("ennoblement.id ASC")
		case domain.EnnoblementSortIDDESC:
			q = q.Order("ennoblement.id DESC")
		case domain.EnnoblementSortServerKeyASC:
			q = q.Order("ennoblement.server_key ASC")
		case domain.EnnoblementSortServerKeyDESC:
			q = q.Order("ennoblement.server_key DESC")
		default:
			return q.Err(fmt.Errorf("%s: %w", s.String(), errInvalidSortValue))
		}
	}

	return q.Limit(a.params.Limit() + 1).Apply(a.applyCursor)
}

func (a listEnnoblementsParamsApplier) applyCursor(q *bun.SelectQuery) *bun.SelectQuery {
	cursor := a.params.Cursor()

	if cursor.IsZero() {
		return q
	}

	sort := a.params.Sort()
	cursorApplier := cursorPaginationApplier{
		data: make([]cursorPaginationApplierDataElement, 0, len(sort)),
	}

	for _, s := range sort {
		var el cursorPaginationApplierDataElement

		switch s {
		case domain.EnnoblementSortIDASC:
			el.value = cursor.ID()
			el.unique = true
			el.column = "ennoblement.id"
			el.direction = sortDirectionASC
		case domain.EnnoblementSortIDDESC:
			el.value = cursor.ID()
			el.unique = true
			el.column = "ennoblement.id"
			el.direction = sortDirectionDESC
		case domain.EnnoblementSortServerKeyASC:
			el.value = cursor.ServerKey()
			el.column = "ennoblement.server_key"
			el.direction = sortDirectionASC
		case domain.EnnoblementSortServerKeyDESC:
			el.value = cursor.ServerKey()
			el.column = "ennoblement.server_key"
			el.direction = sortDirectionDESC
		case domain.EnnoblementSortCreatedAtASC:
			el.value = cursor.CreatedAt()
			el.column = "ennoblement.created_at"
			el.direction = sortDirectionASC
		case domain.EnnoblementSortCreatedAtDESC:
			el.value = cursor.CreatedAt()
			el.column = "ennoblement.created_at"
			el.direction = sortDirectionDESC
		default:
			return q.Err(fmt.Errorf("%s: %w", s.String(), errInvalidSortValue))
		}

		cursorApplier.data = append(cursorApplier.data, el)
	}

	return q.Apply(cursorApplier.apply)
}
