package domain_test

import (
	"fmt"
	"slices"
	"testing"
	"time"

	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"gitea.dwysokinski.me/twhelp/corev3/internal/domain/domaintest"
	"github.com/brianvoe/gofakeit/v7"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestNewCreateEnnoblementParams(t *testing.T) {
	t.Parallel()

	server := domaintest.NewServer(t)

	ennoblements := domain.BaseEnnoblements{
		domaintest.NewBaseEnnoblement(t),
		domaintest.NewBaseEnnoblement(t),
		domaintest.NewBaseEnnoblement(t),
	}

	res, err := domain.NewCreateEnnoblementParams(server.Key(), ennoblements)
	require.NoError(t, err)
	assert.Len(t, res, len(ennoblements))
	for i, e := range ennoblements {
		idx := slices.IndexFunc(res, func(params domain.CreateEnnoblementParams) bool {
			return params.Base().VillageID() == e.VillageID() && params.ServerKey() == server.Key()
		})
		require.GreaterOrEqualf(t, idx, 0, "ennoblements[%d] not found", i)

		params := res[idx]

		assert.Equalf(t, e, params.Base(), "ennoblements[%d]", i)
	}
}

func TestEnnoblementSort_IsInConflict(t *testing.T) {
	t.Parallel()

	type args struct {
		sorts [2]domain.EnnoblementSort
	}

	tests := []struct {
		name        string
		args        args
		expectedRes bool
	}{
		{
			name: "OK: id:ASC serverKey:ASC",
			args: args{
				sorts: [2]domain.EnnoblementSort{domain.EnnoblementSortIDASC, domain.EnnoblementSortServerKeyASC},
			},
			expectedRes: false,
		},
		{
			name: "OK: id:DESC serverKey:ASC",
			args: args{
				sorts: [2]domain.EnnoblementSort{domain.EnnoblementSortIDDESC, domain.EnnoblementSortServerKeyASC},
			},
			expectedRes: false,
		},
		{
			name: "OK: id:ASC id:ASC",
			args: args{
				sorts: [2]domain.EnnoblementSort{domain.EnnoblementSortIDASC, domain.EnnoblementSortIDASC},
			},
			expectedRes: true,
		},
		{
			name: "OK: id:ASC id:DESC",
			args: args{
				sorts: [2]domain.EnnoblementSort{domain.EnnoblementSortIDDESC, domain.EnnoblementSortIDDESC},
			},
			expectedRes: true,
		},
		{
			name: "OK: createdAt:ASC createdAt:DESC",
			args: args{
				sorts: [2]domain.EnnoblementSort{domain.EnnoblementSortCreatedAtDESC, domain.EnnoblementSortCreatedAtDESC},
			},
			expectedRes: true,
		},
		{
			name: "OK: serverKey:DESC serverKey:ASC",
			args: args{
				sorts: [2]domain.EnnoblementSort{domain.EnnoblementSortServerKeyDESC, domain.EnnoblementSortServerKeyASC},
			},
			expectedRes: true,
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			assert.Equal(t, tt.expectedRes, tt.args.sorts[0].IsInConflict(tt.args.sorts[1]))
		})
	}
}

func TestNewEnnoblementCursor(t *testing.T) {
	t.Parallel()

	validEnnoblementCursor := domaintest.NewEnnoblementCursor(t)

	type args struct {
		id        int
		serverKey string
		createdAt time.Time
	}

	type test struct {
		name        string
		args        args
		expectedErr error
	}

	tests := []test{
		{
			name: "OK",
			args: args{
				id:        validEnnoblementCursor.ID(),
				serverKey: validEnnoblementCursor.ServerKey(),
				createdAt: validEnnoblementCursor.CreatedAt(),
			},
			expectedErr: nil,
		},
		{
			name: "ERR: id < 1",
			args: args{
				id:        0,
				serverKey: validEnnoblementCursor.ServerKey(),
				createdAt: validEnnoblementCursor.CreatedAt(),
			},
			expectedErr: domain.ValidationError{
				Model: "EnnoblementCursor",
				Field: "id",
				Err: domain.MinGreaterEqualError{
					Min:     1,
					Current: 0,
				},
			},
		},
	}

	for _, serverKeyTest := range newServerKeyValidationTests() {
		tests = append(tests, test{
			name: serverKeyTest.name,
			args: args{
				id:        validEnnoblementCursor.ID(),
				serverKey: serverKeyTest.key,
			},
			expectedErr: domain.ValidationError{
				Model: "EnnoblementCursor",
				Field: "serverKey",
				Err:   serverKeyTest.expectedErr,
			},
		})
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			ec, err := domain.NewEnnoblementCursor(
				tt.args.id,
				tt.args.serverKey,
				tt.args.createdAt,
			)
			require.ErrorIs(t, err, tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.id, ec.ID())
			assert.Equal(t, tt.args.serverKey, ec.ServerKey())
			assert.Equal(t, tt.args.createdAt, ec.CreatedAt())
			assert.NotEmpty(t, ec.Encode())
		})
	}
}

func TestListEnnoblementsParams_SetServerKeys(t *testing.T) {
	t.Parallel()

	type args struct {
		serverKeys []string
	}

	type test struct {
		name        string
		args        args
		expectedErr error
	}

	tests := []test{
		{
			name: "OK",
			args: args{
				serverKeys: []string{
					domaintest.RandServerKey(),
				},
			},
		},
	}

	for _, serverKeyTest := range newServerKeyValidationTests() {
		tests = append(tests, test{
			name: serverKeyTest.name,
			args: args{
				serverKeys: []string{serverKeyTest.key},
			},
			expectedErr: domain.SliceElementValidationError{
				Model: "ListEnnoblementsParams",
				Field: "serverKeys",
				Index: 0,
				Err:   serverKeyTest.expectedErr,
			},
		})
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListEnnoblementsParams()

			require.ErrorIs(t, params.SetServerKeys(tt.args.serverKeys), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.serverKeys, params.ServerKeys())
		})
	}
}

func TestListEnnoblementsParams_SetSort(t *testing.T) {
	t.Parallel()

	type args struct {
		sort []domain.EnnoblementSort
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				sort: []domain.EnnoblementSort{
					domain.EnnoblementSortCreatedAtASC,
					domain.EnnoblementSortServerKeyASC,
					domain.EnnoblementSortIDASC,
				},
			},
		},
		{
			name: "ERR: len(sort) < 1",
			args: args{
				sort: nil,
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "sort",
				Err: domain.LenOutOfRangeError{
					Min:     1,
					Max:     3,
					Current: 0,
				},
			},
		},
		{
			name: "ERR: len(sort) > 3",
			args: args{
				sort: []domain.EnnoblementSort{
					domain.EnnoblementSortCreatedAtASC,
					domain.EnnoblementSortServerKeyASC,
					domain.EnnoblementSortIDASC,
					domain.EnnoblementSortServerKeyDESC,
				},
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "sort",
				Err: domain.LenOutOfRangeError{
					Min:     1,
					Max:     3,
					Current: 4,
				},
			},
		},
		{
			name: "ERR: conflict",
			args: args{
				sort: []domain.EnnoblementSort{
					domain.EnnoblementSortIDASC,
					domain.EnnoblementSortIDDESC,
				},
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "sort",
				Err: domain.SortConflictError{
					Sort: [2]string{domain.EnnoblementSortIDASC.String(), domain.EnnoblementSortIDDESC.String()},
				},
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListEnnoblementsParams()

			require.ErrorIs(t, params.SetSort(tt.args.sort), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.sort, params.Sort())
		})
	}
}

func TestListEnnoblementsParams_SetEncodedCursor(t *testing.T) {
	t.Parallel()

	validCursor := domaintest.NewEnnoblementCursor(t)

	type args struct {
		cursor string
	}

	tests := []struct {
		name           string
		args           args
		expectedCursor domain.EnnoblementCursor
		expectedErr    error
	}{
		{
			name: "OK",
			args: args{
				cursor: validCursor.Encode(),
			},
			expectedCursor: validCursor,
		},
		{
			name: "ERR: len(cursor) < 1",
			args: args{
				cursor: "",
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "cursor",
				Err: domain.LenOutOfRangeError{
					Min:     1,
					Max:     1000,
					Current: 0,
				},
			},
		},
		{
			name: "ERR: len(cursor) > 1000",
			args: args{
				cursor: gofakeit.LetterN(1001),
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "cursor",
				Err: domain.LenOutOfRangeError{
					Min:     1,
					Max:     1000,
					Current: 1001,
				},
			},
		},
		{
			name: "ERR: malformed base64",
			args: args{
				cursor: "112345",
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "cursor",
				Err:   domain.ErrInvalidCursor,
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListEnnoblementsParams()

			require.ErrorIs(t, params.SetEncodedCursor(tt.args.cursor), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.cursor, params.Cursor().Encode())
		})
	}
}

func TestListEnnoblementsParams_SetLimit(t *testing.T) {
	t.Parallel()

	type args struct {
		limit int
	}

	tests := []struct {
		name        string
		args        args
		expectedErr error
	}{
		{
			name: "OK",
			args: args{
				limit: domain.EnnoblementListMaxLimit,
			},
		},
		{
			name: "ERR: limit < 1",
			args: args{
				limit: 0,
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "limit",
				Err: domain.MinGreaterEqualError{
					Min:     1,
					Current: 0,
				},
			},
		},
		{
			name: fmt.Sprintf("ERR: limit > %d", domain.EnnoblementListMaxLimit),
			args: args{
				limit: domain.EnnoblementListMaxLimit + 1,
			},
			expectedErr: domain.ValidationError{
				Model: "ListEnnoblementsParams",
				Field: "limit",
				Err: domain.MaxLessEqualError{
					Max:     domain.EnnoblementListMaxLimit,
					Current: domain.EnnoblementListMaxLimit + 1,
				},
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			t.Parallel()

			params := domain.NewListEnnoblementsParams()

			require.ErrorIs(t, params.SetLimit(tt.args.limit), tt.expectedErr)
			if tt.expectedErr != nil {
				return
			}
			assert.Equal(t, tt.args.limit, params.Limit())
		})
	}
}

func TestNewListEnnoblementsResult(t *testing.T) {
	t.Parallel()

	ennoblements := domain.Ennoblements{
		domaintest.NewEnnoblement(t),
		domaintest.NewEnnoblement(t),
		domaintest.NewEnnoblement(t),
	}
	next := domaintest.NewEnnoblement(t)

	t.Run("OK: with next", func(t *testing.T) {
		t.Parallel()

		res, err := domain.NewListEnnoblementsResult(ennoblements, next)
		require.NoError(t, err)
		assert.Equal(t, ennoblements, res.Ennoblements())
		assert.Equal(t, ennoblements[0].ID(), res.Self().ID())
		assert.Equal(t, ennoblements[0].ServerKey(), res.Self().ServerKey())
		assert.Equal(t, ennoblements[0].CreatedAt(), res.Self().CreatedAt())
		assert.Equal(t, next.ID(), res.Next().ID())
		assert.Equal(t, next.ServerKey(), res.Next().ServerKey())
		assert.Equal(t, next.CreatedAt(), res.Next().CreatedAt())
	})

	t.Run("OK: without next", func(t *testing.T) {
		t.Parallel()

		res, err := domain.NewListEnnoblementsResult(ennoblements, domain.Ennoblement{})
		require.NoError(t, err)
		assert.Equal(t, ennoblements, res.Ennoblements())
		assert.Equal(t, ennoblements[0].ID(), res.Self().ID())
		assert.Equal(t, ennoblements[0].ServerKey(), res.Self().ServerKey())
		assert.Equal(t, ennoblements[0].CreatedAt(), res.Self().CreatedAt())
		assert.True(t, res.Next().IsZero())
	})

	t.Run("OK: 0 ennoblements", func(t *testing.T) {
		t.Parallel()

		res, err := domain.NewListEnnoblementsResult(nil, domain.Ennoblement{})
		require.NoError(t, err)
		assert.Zero(t, res.Ennoblements())
		assert.True(t, res.Self().IsZero())
		assert.True(t, res.Next().IsZero())
	})
}

func TestNewListEnnoblementsWithRelationsResult(t *testing.T) {
	t.Parallel()

	ennoblements := domain.EnnoblementsWithRelations{
		domaintest.NewEnnoblementWithRelations(t),
		domaintest.NewEnnoblementWithRelations(t),
		domaintest.NewEnnoblementWithRelations(t),
	}
	next := domaintest.NewEnnoblementWithRelations(t)

	t.Run("OK: with next", func(t *testing.T) {
		t.Parallel()

		res, err := domain.NewListEnnoblementsWithRelationsResult(ennoblements, next)
		require.NoError(t, err)
		assert.Equal(t, ennoblements, res.Ennoblements())
		assert.Equal(t, ennoblements[0].Ennoblement().ID(), res.Self().ID())
		assert.Equal(t, ennoblements[0].Ennoblement().ServerKey(), res.Self().ServerKey())
		assert.Equal(t, ennoblements[0].Ennoblement().CreatedAt(), res.Self().CreatedAt())
		assert.Equal(t, next.Ennoblement().ID(), res.Next().ID())
		assert.Equal(t, next.Ennoblement().ServerKey(), res.Next().ServerKey())
		assert.Equal(t, next.Ennoblement().CreatedAt(), res.Next().CreatedAt())
	})

	t.Run("OK: without next", func(t *testing.T) {
		t.Parallel()

		res, err := domain.NewListEnnoblementsWithRelationsResult(ennoblements, domain.EnnoblementWithRelations{})
		require.NoError(t, err)
		assert.Equal(t, ennoblements, res.Ennoblements())
		assert.Equal(t, ennoblements[0].Ennoblement().ID(), res.Self().ID())
		assert.Equal(t, ennoblements[0].Ennoblement().ServerKey(), res.Self().ServerKey())
		assert.Equal(t, ennoblements[0].Ennoblement().CreatedAt(), res.Self().CreatedAt())
		assert.True(t, res.Next().IsZero())
	})

	t.Run("OK: 0 ennoblements", func(t *testing.T) {
		t.Parallel()

		res, err := domain.NewListPlayersWithRelationsResult(nil, domain.PlayerWithRelations{})
		require.NoError(t, err)
		assert.Zero(t, res.Players())
		assert.True(t, res.Self().IsZero())
		assert.True(t, res.Next().IsZero())
	})
}
